use super::provision;
use super::Keyfork;
use crate::{clap_ext::*, config};
use clap::{builder::PossibleValue, Parser, Subcommand, ValueEnum};
use std::{
    collections::HashMap,
    fmt::Display,
    fs::File,
    io::Write,
    path::{Path, PathBuf},
    str::FromStr,
};

use keyfork_derive_openpgp::{
    openpgp::{
        self,
        armor::{Kind, Writer},
        packet::UserID,
        policy::StandardPolicy,
        serialize::{
            stream::{Encryptor2, LiteralWriter, Message, Recipient},
            Serialize,
        },
        types::KeyFlags,
    },
    XPrv,
};
use keyfork_prompt::default_handler;
use keyfork_shard::{openpgp::OpenPGP, Format};

type StringMap = HashMap<String, String>;

#[derive(Clone, Debug, Default)]
pub enum SeedSize {
    Bits128,

    #[default]
    Bits256,
}

// custom impl to override names in ValueEnum
impl ValueEnum for SeedSize {
    fn value_variants<'a>() -> &'a [Self] {
        &[Self::Bits128, Self::Bits256]
    }

    fn to_possible_value(&self) -> Option<PossibleValue> {
        Some(PossibleValue::new(match self {
            SeedSize::Bits128 => "128",
            SeedSize::Bits256 => "256",
        }))
    }
}

#[derive(thiserror::Error, Debug, Clone)]
pub enum SeedSizeError {
    #[error("Expected one of 128, 256")]
    InvalidChoice,
}

impl Display for SeedSize {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            SeedSize::Bits128 => write!(f, "128"),
            SeedSize::Bits256 => write!(f, "256"),
        }
    }
}

impl std::str::FromStr for SeedSize {
    type Err = SeedSizeError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "128" => SeedSize::Bits128,
            "256" => SeedSize::Bits256,
            _ => return Err(SeedSizeError::InvalidChoice),
        })
    }
}

impl From<&SeedSize> for usize {
    fn from(value: &SeedSize) -> Self {
        match value {
            SeedSize::Bits128 => 128,
            SeedSize::Bits256 => 256,
        }
    }
}
#[derive(Clone, Debug, thiserror::Error)]
pub enum MnemonicSeedSourceParseError {
    #[error("Expected one of system, playing, tarot, dice")]
    InvalidChoice,
}

#[derive(Clone, Debug, Default, ValueEnum)]
pub enum MnemonicSeedSource {
    /// System entropy
    #[default]
    System,

    /// Playing cards
    Playing,

    /// Tarot cards
    Tarot,

    /// Dice
    Dice,
}

impl std::str::FromStr for MnemonicSeedSource {
    type Err = MnemonicSeedSourceParseError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(match s {
            "system" => Self::System,
            "playing" => Self::Playing,
            "tarot" => Self::Tarot,
            "dice" => Self::Dice,
            _ => return Err(Self::Err::InvalidChoice),
        })
    }
}

impl MnemonicSeedSource {
    pub fn handle(
        &self,
        size: &SeedSize,
    ) -> Result<keyfork_mnemonic::Mnemonic, Box<dyn std::error::Error>> {
        let size = match size {
            SeedSize::Bits128 => 128,
            SeedSize::Bits256 => 256,
        };
        let seed = match self {
            MnemonicSeedSource::System => keyfork_entropy::generate_entropy_of_size(size / 8)?,
            MnemonicSeedSource::Playing => todo!(),
            MnemonicSeedSource::Tarot => todo!(),
            MnemonicSeedSource::Dice => todo!(),
        };
        let mnemonic = keyfork_mnemonic::Mnemonic::try_from_slice(&seed)?;
        Ok(mnemonic)
    }
}

#[derive(Subcommand, Clone, Debug)]
pub enum MnemonicSubcommands {
    /// Generate a mnemonic using a given entropy source.
    ///
    /// Mnemonics are a form of encoding a given form of entropy and are used to create a master
    /// seed for BIP-0032 hierarchial derivation. The mnemonic is like the "password" for all of
    /// Keyfork's derivations, and should be treated securely. This command provides a secure
    /// method of generating a seed using system entropy, as well as various forms of loading
    /// physicalized entropy into a mnemonic. The mnemonic should be stored in a safe location
    /// (such as a Trezor "recovery seed card") and never persisted digitally.
    ///
    /// When using the `--shard`, `--shard-to`, `--encrypt-to`, and `--encrypt-to-self` +
    /// `--provision` arguments, the mnemonic is _not_ sent to output. The data for the mnemonic is
    /// then either split using Keyfork Shard or encrypted using OpenPGP.
    Generate {
        /// The source from where a seed is created.
        #[arg(long, value_enum, default_value_t = Default::default())]
        source: MnemonicSeedSource,

        /// The size of the mnemonic, in bits.
        #[arg(long, default_value_t = Default::default())]
        size: SeedSize,

        /// Encrypt the mnemonic to an OpenPGP certificate in the provided path.
        ///
        /// When given arguments in the format `--encrypt-to input.asc,output=output.asc`, the
        /// output of the encryption will be written to `output.asc`. Otherwise, the default
        /// behavior is to write the output to `input.enc.asc`. If the output file already exists,
        /// it will not be overwritten, and the command will exit unsuccessfully.
        #[arg(long)]
        encrypt_to: Option<Vec<ValueWithOptions<PathBuf>>>,

        /// Shard the mnemonic to the certificates in the given Shardfile. Requires a decrypt
        /// operation on the Shardfile to access the metadata and certificates.
        ///
        /// When given arguments in the format `--encrypt-to input.asc,output=output.asc`, the
        /// output of the encryption will be written to `output.asc`. Otherwise, the default
        /// behavior is to write the output to `input.new.asc`. If the output file already exists,
        /// it will not be overwritten, and the command will exit unsuccessfully.
        #[arg(long)]
        shard_to: Option<Vec<ValueWithOptions<PathBuf>>>,

        /// Shard the mnemonic to the provided certificates.
        ///
        /// The following additional arguments are available:
        ///
        /// * threshold, m: the minimum amount of shares required to reconstitute the shard. By
        ///   default, this is the amount of certificates provided.
        ///
        /// * max, n: the maximum amount of shares. When provided, this is used to ensure the
        ///   certificate count is correct. This is required when using `threshold` or `m`.
        ///
        /// * output: the file to write the generated Shardfile to. By default, assuming the
        ///   certificate input is `input.asc`, the generated Shardfile would be written to
        ///   `input.shard.asc`.
        #[arg(long)]
        shard: Option<Vec<ValueWithOptions<PathBuf>>>,

        /// Encrypt the mnemonic to an OpenPGP certificate derived from the mnemonic, writing the
        /// output to the provided path. This command must be run in combination with
        /// `--provision openpgp-card` or another relevant provisioner, to ensure the newly
        /// generated mnemonic would be decryptable by some form of provisioned hardware.
        ///
        /// When given arguments in the format `--encrypt-to-self output.asc,output=encrypted.asc`,
        /// the output of the OpenPGP certificate will be written to `output.asc`, while the output
        /// of the encryption will be written to `encrypted.asc`. Otherwise, the
        /// default behavior is to write the encrypted mnemonic to `output.enc.asc`. If either
        /// output file already exists, it will not be overwritten, and the command will exit
        /// unsuccessfully.
        ///
        /// Additionally, when given the `account=` option (which must match the `account=` option
        /// of the relevant provisioner), the given account will be used instead of the default
        /// account of 0.
        ///
        /// Because a new OpenPGP key needs to be created, a User ID can also be supplied, using
        /// the option `userid=<your User ID>`. It can contain any characters that are not a comma.
        #[arg(long)]
        encrypt_to_self: Option<ValueWithOptions<PathBuf>>,

        /// Provision a key derived from the mnemonic to a piece of hardware such as an OpenPGP
        /// smartcard. This argument is required when used with `--encrypt-to-self`.
        ///
        /// Additional arguments, such as the amount of hardware to provision and the
        /// account to use when deriving, can be specified by using (for example)
        /// `--provision openpgp-card,count=2,account=1`.
        #[arg(long)]
        provision: Option<ValueWithOptions<provision::Provisioner>>,
    },
}

// NOTE: This function defaults to `.asc` in the event no extension is found.
// This is specific to OpenPGP. If you want to use this function elsewhere (why?),
// be sure to use a relevant extension for your context.
fn determine_valid_output_path<T: AsRef<Path>>(
    path: &Path,
    mid_ext: &str,
    optional_path: Option<T>,
) -> PathBuf {
    match optional_path {
        Some(p) => p.as_ref().to_path_buf(),
        None => {
            let extension = match path.extension() {
                Some(ext) => format!("{mid_ext}.{ext}", ext = ext.to_string_lossy()),
                None => format!("{mid_ext}.asc"),
            };
            path.with_extension(extension)
        }
    }
}

fn is_extension_armored(path: &Path) -> bool {
    match path.extension().and_then(|s| s.to_str()) {
        Some("pgp") | Some("gpg") => false,
        Some("asc") => true,
        _ => {
            eprintln!("unable to determine whether to armor file: {path:?}");
            eprintln!("use .gpg, .pgp, or .asc extension, or `armor=true`");
            eprintln!("defaulting to armored");
            true
        }
    }
}

fn do_encrypt_to(
    mnemonic: &keyfork_mnemonic::Mnemonic,
    path: &Path,
    options: &StringMap,
) -> Result<(), Box<dyn std::error::Error>> {
    let policy = StandardPolicy::new();

    let output_file = determine_valid_output_path(path, "enc", options.get("output"));

    let is_armored =
        options.get("armor").is_some_and(|a| a == "true") || is_extension_armored(&output_file);

    let certs = OpenPGP::discover_certs(path)?;
    let valid_certs = certs
        .iter()
        .map(|c| c.with_policy(&policy, None))
        .collect::<openpgp::Result<Vec<_>>>()?;
    let recipients = valid_certs.iter().flat_map(|valid_cert| {
        let keys = valid_cert.keys().alive().for_storage_encryption();
        keys.map(|key| Recipient::new(key.keyid(), key.key()))
    });

    let mut output = vec![];
    let message = Message::new(&mut output);
    let encrypted_message = Encryptor2::for_recipients(message, recipients).build()?;
    let mut literal_message = LiteralWriter::new(encrypted_message).build()?;
    literal_message.write_all(mnemonic.to_string().as_bytes())?;
    literal_message.write_all(b"\n")?;
    literal_message.finalize()?;

    let mut file = File::create_new(&output_file)?;
    if is_armored {
        let mut writer = Writer::new(file, Kind::Message)?;
        writer.write_all(&output)?;
        writer.finalize()?;
    } else {
        file.write_all(&output)?;
    }

    Ok(())
}

fn do_encrypt_to_self(
    mnemonic: &keyfork_mnemonic::Mnemonic,
    path: &Path,
    options: &StringMap,
) -> Result<(), Box<dyn std::error::Error>> {
    let output_file = determine_valid_output_path(path, "enc", options.get("output"));

    let is_armored =
        options.get("armor").is_some_and(|a| a == "true") || is_extension_armored(&output_file);

    let account = options
        .get("account")
        .map(|account| u32::from_str(account))
        .transpose()?
        .unwrap_or(0);
    let account_index = keyfork_derive_util::DerivationIndex::new(account, true)?;

    let userid = options
        .get("userid")
        .map(|userid| UserID::from(userid.as_str()));

    let subkeys = [
        KeyFlags::empty().set_certification(),
        KeyFlags::empty().set_signing(),
        KeyFlags::empty()
            .set_transport_encryption()
            .set_storage_encryption(),
        KeyFlags::empty().set_authentication(),
    ];

    let seed = mnemonic.generate_seed(None);
    let xprv = XPrv::new(seed)?;
    let derivation_path = keyfork_derive_path_data::paths::OPENPGP
        .clone()
        .chain_push(account_index);

    let cert = keyfork_derive_openpgp::derive(
        xprv.derive_path(&derivation_path)?,
        &subkeys,
        &userid.unwrap_or(UserID::from("Keyfork-Generated Key")),
    )?;

    let mut file = File::create_new(path)?;
    if is_armored {
        let mut writer = Writer::new(file, Kind::PublicKey)?;
        cert.serialize(&mut writer)?;
        writer.finalize()?;
    } else {
        cert.serialize(&mut file)?;
    }

    // a sneaky bit of DRY
    do_encrypt_to(
        mnemonic,
        path,
        &StringMap::from([(
            String::from("output"),
            output_file.to_string_lossy().to_string(),
        )]),
    )?;

    Ok(())
}

#[derive(thiserror::Error, Debug)]
#[error("Either the threshold(m) or the max(n) values are missing")]
struct MissingThresholdOrMax;

fn do_shard(
    mnemonic: &keyfork_mnemonic::Mnemonic,
    path: &Path,
    options: &StringMap,
) -> Result<(), Box<dyn std::error::Error>> {
    let output_file = determine_valid_output_path(path, "shard", options.get("output"));

    let is_armored =
        options.get("armor").is_some_and(|a| a == "true") || is_extension_armored(&output_file);

    let threshold = options
        .get("threshold")
        .or_else(|| options.get("m"))
        .map(|s| u8::from_str(s))
        .transpose()?;

    let max = options
        .get("max")
        .or_else(|| options.get("n"))
        .map(|s| u8::from_str(s))
        .transpose()?;

    let certs = OpenPGP::discover_certs(path)?;

    // if neither are set: false
    // if both are set: false
    // if only one is set: true

    if threshold.is_some() ^ max.is_some() {
        return Err(MissingThresholdOrMax)?;
    }

    let (threshold, max) = match threshold.zip(max) {
        Some(t) => t,
        None => {
            let len = u8::try_from(certs.len())?;
            (len, len)
        }
    };

    let openpgp = keyfork_shard::openpgp::OpenPGP;

    let mut output = vec![];
    openpgp.shard_and_encrypt(threshold, max, mnemonic.as_bytes(), &certs[..], &mut output)?;

    let mut file = File::create_new(&output_file)?;
    if is_armored {
        file.write_all(&output)?;
    } else {
        todo!("keyfork does not handle binary shardfiles");
        /*
         * NOTE: this code works, but can't be recombined by Keyfork.
         * therefore, we'll error, before someone tries to use it.
        let mut dearmor = Reader::from_bytes(&output, ReaderMode::Tolerant(None));
        std::io::copy(&mut dearmor, &mut file)?;
        */
    }

    Ok(())
}

fn do_shard_to(
    mnemonic: &keyfork_mnemonic::Mnemonic,
    path: &Path,
    options: &StringMap,
) -> Result<(), Box<dyn std::error::Error>> {
    let output_file = determine_valid_output_path(path, "new", options.get("output"));

    let is_armored =
        options.get("armor").is_some_and(|a| a == "true") || is_extension_armored(&output_file);

    let openpgp = keyfork_shard::openpgp::OpenPGP;
    let prompt = default_handler()?;

    let input = File::open(path)?;
    let (threshold, certs) = openpgp.decrypt_metadata_from_file(
        Some(&[][..]), // the things i must do to avoid qualifying types.
        input,
        prompt,
    )?;

    let mut output = vec![];
    openpgp.shard_and_encrypt(
        threshold,
        u8::try_from(certs.len())?,
        mnemonic.as_bytes(),
        &certs[..],
        &mut output,
    )?;

    let mut file = File::create_new(&output_file)?;
    if is_armored {
        file.write_all(&output)?;
    } else {
        todo!("keyfork does not handle binary shardfiles");
        /*
         * NOTE: this code works, but can't be recombined by Keyfork.
         * therefore, we'll error, before someone tries to use it.
        let mut dearmor = Reader::from_bytes(&output, ReaderMode::Tolerant(None));
        std::io::copy(&mut dearmor, &mut file)?;
        */
    }

    Ok(())
}

#[derive(thiserror::Error, Debug)]
#[error("missing key: {0}")]
struct MissingKey(&'static str);

fn do_provision(
    mnemonic: &keyfork_mnemonic::Mnemonic,
    provisioner: &provision::Provisioner,
    options: &StringMap,
) -> Result<(), Box<dyn std::error::Error>> {
    let mut options = options.clone();
    let account = options
        .remove("account")
        .map(|account| u32::from_str(&account))
        .transpose()?
        .unwrap_or(0);
    let identifier = options
        .remove("identifier")
        .ok_or(MissingKey("identifier"))?
        .split(',')
        .map(String::from)
        .collect::<Vec<_>>();
    let count = options
        .remove("count")
        .map(|count| usize::from_str(&count))
        .transpose()?
        .unwrap_or(identifier.len());

    for (_, identifier) in (0..count).zip(identifier.into_iter()) {
        let provisioner_config = config::Provisioner {
            account,
            identifier,
            metadata: Some(options.clone()),
        };

        provisioner.provision_with_mnemonic(mnemonic, provisioner_config.clone())?;
    }

    Ok(())
}

impl MnemonicSubcommands {
    pub fn handle(
        &self,
        _m: &Mnemonic,
        _keyfork: &Keyfork,
    ) -> Result<(), Box<dyn std::error::Error>> {
        match self {
            MnemonicSubcommands::Generate {
                source,
                size,
                encrypt_to,
                shard_to,
                shard,
                encrypt_to_self,
                provision,
            } => {
                // NOTE: We should never have a case where there's Some() of empty vec, but
                // we will make sure to check it just in case.
                let mut will_print_mnemonic =
                    encrypt_to.is_none() || encrypt_to.as_ref().is_some_and(|e| e.is_empty());
                will_print_mnemonic = will_print_mnemonic && shard_to.is_none()
                    || shard_to.as_ref().is_some_and(|s| s.is_empty());
                will_print_mnemonic = will_print_mnemonic && shard.is_none()
                    || shard.as_ref().is_some_and(|s| s.is_empty());
                will_print_mnemonic = will_print_mnemonic
                    && (encrypt_to_self.as_ref().is_none() || provision.as_ref().is_none());

                let mnemonic = source.handle(size)?;

                if let Some(encrypt_to) = encrypt_to {
                    for entry in encrypt_to {
                        do_encrypt_to(&mnemonic, &entry.inner, &entry.values)?;
                    }
                }

                if let Some(encrypt_to_self) = encrypt_to_self {
                    do_encrypt_to_self(&mnemonic, &encrypt_to_self.inner, &encrypt_to_self.values)?;
                }

                if let Some(provisioner) = provision {
                    do_provision(&mnemonic, &provisioner.inner, &provisioner.values)?;
                }

                if let Some(shard_to) = shard_to {
                    for entry in shard_to {
                        do_shard_to(&mnemonic, &entry.inner, &entry.values)?;
                    }
                }

                if let Some(shard) = shard {
                    for entry in shard {
                        do_shard(&mnemonic, &entry.inner, &entry.values)?;
                    }
                }

                if will_print_mnemonic {
                    println!("{}", mnemonic);
                }
                Ok(())
            }
        }
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Mnemonic {
    #[command(subcommand)]
    pub command: MnemonicSubcommands,
}
