use super::Keyfork;
use crate::config;

use clap::{builder::PossibleValue, Parser, Subcommand, ValueEnum};

use keyfork_derive_util::{DerivationIndex, ExtendedPrivateKey};

mod openpgp;

type Identifier = (String, Option<String>);

#[derive(Debug, Clone)]
pub enum Provisioner {
    OpenPGPCard(openpgp::OpenPGPCard),
}

impl std::fmt::Display for Provisioner {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Provisioner::OpenPGPCard(_) => f.write_str("openpgp-card"),
        }
    }
}

impl Provisioner {
    pub fn discover(&self) -> Result<Vec<Identifier>, Box<dyn std::error::Error>> {
        match self {
            Provisioner::OpenPGPCard(o) => o.discover(),
        }
    }

    pub fn provision(
        &self,
        provisioner: config::Provisioner,
    ) -> Result<(), Box<dyn std::error::Error>> {
        match self {
            Provisioner::OpenPGPCard(o) => {
                type Prv = <openpgp::OpenPGPCard as ProvisionExec>::PrivateKey;
                type XPrv = ExtendedPrivateKey<Prv>;
                let account_index = DerivationIndex::new(provisioner.account, true)?;
                let path = <openpgp::OpenPGPCard as ProvisionExec>::derivation_prefix()
                    .chain_push(account_index);
                let mut client = keyforkd_client::Client::discover_socket()?;
                let xprv: XPrv = client.request_xprv(&path)?;
                o.provision(xprv, provisioner)
            }
        }
    }

    pub fn provision_with_mnemonic(
        &self,
        mnemonic: &keyfork_mnemonic::Mnemonic,
        provisioner: config::Provisioner,
    ) -> Result<(), Box<dyn std::error::Error>> {
        match self {
            Provisioner::OpenPGPCard(o) => {
                type Prv = <openpgp::OpenPGPCard as ProvisionExec>::PrivateKey;
                type XPrv = ExtendedPrivateKey<Prv>;
                let account_index = DerivationIndex::new(provisioner.account, true)?;
                let path = <openpgp::OpenPGPCard as ProvisionExec>::derivation_prefix()
                    .chain_push(account_index);
                let xprv = XPrv::new(mnemonic.generate_seed(None))?.derive_path(&path)?;
                o.provision(xprv, provisioner)
            }
        }
    }
}

impl ValueEnum for Provisioner {
    fn value_variants<'a>() -> &'a [Self] {
        &[Self::OpenPGPCard(openpgp::OpenPGPCard)]
    }

    fn to_possible_value(&self) -> Option<PossibleValue> {
        Some(PossibleValue::new(match self {
            Self::OpenPGPCard(_) => "openpgp-card",
        }))
    }
}

#[derive(Debug, thiserror::Error)]
#[error("The given value could not be matched as a provisioner: {0} ({1})")]
pub struct ProvisionerFromStrError(String, String);

impl std::str::FromStr for Provisioner {
    type Err = ProvisionerFromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        <Provisioner as ValueEnum>::from_str(s, false)
            .map_err(|e| ProvisionerFromStrError(s.to_string(), e))
    }
}

trait ProvisionExec {
    type PrivateKey: keyfork_derive_util::PrivateKey + Clone;

    /// Discover all known places the formatted key can be deployed to.
    fn discover(&self) -> Result<Vec<Identifier>, Box<dyn std::error::Error>> {
        Ok(vec![])
    }

    /// Return the derivation path for deriving keys.
    fn derivation_prefix() -> keyfork_derive_util::DerivationPath;

    /// Derive a key and deploy it to a target.
    fn provision(
        &self,
        xprv: keyfork_derive_util::ExtendedPrivateKey<Self::PrivateKey>,
        p: config::Provisioner,
    ) -> Result<(), Box<dyn std::error::Error>>;
}

#[derive(Subcommand, Clone, Debug)]
pub enum ProvisionSubcommands {
    /// Discover all available targets on the system.
    Discover,
}

// NOTE: All struct fields are marked as Option so they may be constructed when running a
// subcommand. This is allowed through marking the parent command subcommand_negates_reqs(true).

#[derive(Parser, Debug, Clone)]
pub struct Provision {
    #[command(subcommand)]
    pub subcommand: Option<ProvisionSubcommands>,

    provisioner_name: Provisioner,

    /// Account ID.
    #[arg(long, required(true))]
    account_id: Option<u32>,

    /// Identifier of the hardware to deploy to, listable by running the `discover` subcommand.
    #[arg(long, required(true))]
    identifier: Option<String>,
}

// NOTE: In the future, this impl will be used by `keyfork recover` to reprovision hardware from
// the config file, iterating through a [Config::provisioner].
// TODO: How can metadata be passed in?

#[derive(thiserror::Error, Debug)]
#[error("Missing field: {0}")]
pub struct MissingField(&'static str);

impl TryFrom<Provision> for config::Provisioner {
    type Error = MissingField;

    fn try_from(value: Provision) -> Result<Self, Self::Error> {
        Ok(Self {
            account: value.account_id.ok_or(MissingField("account_id"))?,
            identifier: value.identifier.ok_or(MissingField("identifier"))?,
            metadata: Default::default(),
        })
    }
}

impl Provision {
    pub fn handle(&self, _keyfork: &Keyfork) -> Result<(), Box<dyn std::error::Error>> {
        match self.subcommand {
            Some(ProvisionSubcommands::Discover) => {
                let mut iter = self.provisioner_name.discover()?.into_iter().peekable();
                while let Some((identifier, context)) = iter.next() {
                    println!("Identifier: {identifier}");
                    if let Some(context) = context {
                        println!("Context: {context}");
                    }
                    if iter.peek().is_some() {
                        println!("---");
                    }
                }
            }
            None => {
                self.provisioner_name.provision(self.clone().try_into()?)?;
            }
        }
        Ok(())
    }
}
