use super::ProvisionExec;
use crate::{config, openpgp_card::factory_reset_current_card};

use card_backend_pcsc::PcscBackend;
use keyfork_derive_openpgp::{
    openpgp::{packet::UserID, types::KeyFlags},
    XPrv,
};
use keyfork_prompt::{
    default_handler, prompt_validated_passphrase,
    validators::{SecurePinValidator, Validator},
};
use openpgp_card_sequoia::{state::Open, Card};

#[derive(Clone, Debug)]
pub struct OpenPGPCard;

#[derive(thiserror::Error, Debug)]
#[error("Provisioner was unable to find a matching smartcard")]
struct NoMatchingSmartcard;

impl ProvisionExec for OpenPGPCard {
    type PrivateKey = keyfork_derive_openpgp::XPrvKey;

    fn discover(&self) -> Result<Vec<(String, Option<String>)>, Box<dyn std::error::Error>> {
        let mut idents = vec![];
        for backend in PcscBackend::cards(None)? {
            let backend = backend?;
            let mut card = Card::<Open>::new(backend)?;
            let mut transaction = card.transaction()?;
            let identifier = transaction.application_identifier()?.ident();
            let name = transaction.cardholder_name()?;
            let name = (!name.is_empty()).then_some(name);
            idents.push((identifier, name));
        }
        Ok(idents)
    }

    fn derivation_prefix() -> keyfork_derive_util::DerivationPath {
        keyfork_derive_path_data::paths::OPENPGP.clone()
    }

    fn provision(
        &self,
        xprv: XPrv,
        provisioner: config::Provisioner,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let mut pm = default_handler()?;
        let user_pin_validator = SecurePinValidator {
            min_length: Some(6),
            ..Default::default()
        }
        .to_fn();
        let admin_pin_validator = SecurePinValidator {
            min_length: Some(8),
            ..Default::default()
        }
        .to_fn();

        let user_pin = prompt_validated_passphrase(
            &mut *pm,
            "Please enter the new smartcard User PIN: ",
            3,
            &user_pin_validator,
        )?;
        let admin_pin = prompt_validated_passphrase(
            &mut *pm,
            "Please enter the new smartcard Admin PIN: ",
            3,
            &admin_pin_validator,
        )?;

        let mut has_provisioned = false;

        for backend in PcscBackend::cards(None)? {
            let backend = backend?;

            let subkeys = vec![
                KeyFlags::empty().set_certification(),
                KeyFlags::empty().set_signing(),
                KeyFlags::empty()
                    .set_transport_encryption()
                    .set_storage_encryption(),
                KeyFlags::empty().set_authentication(),
            ];

            // NOTE: This User ID doesn't have meaningful context on the card.
            // To give it a reasonable name, use `keyfork derive openpgp` or some other system that
            // generates the OpenPGP certificate.
            let userid = UserID::from("Keyfork-Provisioned Key");
            let cert = keyfork_derive_openpgp::derive(xprv.clone(), &subkeys, &userid)?;

            let result = factory_reset_current_card(
                &mut |identifier| { identifier == provisioner.identifier },
                user_pin.trim(),
                admin_pin.trim(),
                &cert,
                &keyfork_derive_openpgp::openpgp::policy::StandardPolicy::new(),
                backend,
            )?;

            has_provisioned = has_provisioned || result;
        }

        if !has_provisioned {
            return Err(NoMatchingSmartcard)?;
        }

        Ok(())
    }
}
